/*
 *                            *
 *                           * *
 *                          *   *
 *                     ***************
 *                      * *       * *
 *                       *  MUMPS  *
 *                      * *       * *
 *                     ***************
 *                          *   *
 *                           * *
 *                            *
 *
 *   transact.c
 *    FreeM transaction processing support
 *
 *  
 *   Author: John P. Willis <jpw@coherent-logic.com>
 *    Copyright (C) 1998 MUG Deutschland
 *    Copyright (C) 2020 Coherent Logic Development LLC
 *
 *   Last modified: 16 Mar 2020
 * 
 **/

#include <string.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>

#include "mpsdef.h"
#include "transact.h"
#include "iftab.h"
#include "journal.h"

#define FALSE   0
#define TRUE    1


int tp_committing = FALSE;
int tp_level = 0;
tp_transaction transactions[TP_MAX_NEST];


int tp_tstart(char *tp_id, short serial, short restartable, char **sym_save)
{
    if (tp_level == TP_MAX_NEST) {
        char m[256];

        snprintf (m, 256, "Attempt to exceed TP_MAX_NEST. Transaction aborted.\r\n\201");
        write_m (m);

        return FALSE;
    }

	tp_level++;

    jnl_ent_write (JNLA_TSTART, "", "");
    
    strcpy (transactions[tp_level].tp_id, tp_id);

    transactions[tp_level].serial = serial;
    transactions[tp_level].restartable = restartable;

    transactions[tp_level].opcount = 0;

    return TRUE;

}

int tp_add_op(short islock, short action, char *key, char *data)
{
    int oc;

    if (transactions[tp_level].opcount == TP_MAX_OPS) {
        char m[256];

        snprintf (m, 256, "Attempt to exceed TP_MAX_OPS at transaction level %d. Rolling back.\r\n\201", tp_level);
        write_m (m);

        tp_trollback (1);

        return FALSE;
    }

    /* update transaction-in-flight symbol table */
    switch (action) {

        case set_sym:
            iftab_insert (key, data, tp_level);
            break;

        case kill_sym:
            iftab_kill (key);
            break;
            
    }

    transactions[tp_level].opcount = transactions[tp_level].opcount + 1;

    oc = transactions[tp_level].opcount;

    transactions[tp_level].ops[oc].is_lock = islock;
    transactions[tp_level].ops[oc].action = action;

    stcpy (&transactions[tp_level].ops[oc].key, key);
    stcpy (&transactions[tp_level].ops[oc].data, data);
       
 
    return TRUE;
}

int tp_tcommit(void)
{
    register int i;

    tp_committing = TRUE;
    
    for(i = 1; i <= transactions[tp_level].opcount; i++) {

        if (transactions[tp_level].ops[i].is_lock == FALSE) {
            global (transactions[tp_level].ops[i].action, transactions[tp_level].ops[i].key, transactions[tp_level].ops[i].data);
        }

    }

	tp_trollback (1);

    tp_committing = FALSE;

    return TRUE;
}

int tp_trollback(int levels)
{
    register int i;

    for(i = 1; i <= transactions[tp_level].opcount; i++) {

        if (transactions[tp_level].ops[i].is_lock == TRUE) {

            /* TODO: this is a lock. release it. */
            
        }

    }

    for (i = tp_level; i >= (((tp_level - levels) >= 0) ? tp_level - levels : 0); i--) {
        iftab_pop_tlevel (i);
    }

    tp_level = ((tp_level - levels) >= 0) ? tp_level - levels : 0;

    return TRUE;
}

int tp_trestart(void)
{

}

void tp_tdump(void)
{
    tp_gblop *op;
    int i, j;

    char tkey[256];
    char tdata[256];

    set_io (UNIX);

    if (tp_level == 0) {
        printf("No transaction is active.\n");

        return;
    }

    for(i = 1; i <= tp_level; i++) {

        if(i == tp_level) {
            printf(" $TLEVEL %d*\n", i);
        }
        else {
            printf(" $TLEVEL %d\n", i);
        }

        printf("  Operations for Transaction ID: %s\n", transactions[i].tp_id);

        for(j = 1; j <= transactions[i].opcount; j++) {
            printf("  %d:\t", j);
            stcpy (tkey, transactions[i].ops[j].key);
            stcnv_m2c (tkey);
            stcpy (tdata, transactions[i].ops[j].data);
            stcnv_m2c (tdata);
            
            printf("action = %d  key = %s  data = %s\n", transactions[i].ops[j].action, tkey, tdata);
        }

    }


    set_io (MUMPS);
}